from httpx import AsyncClient
from opentelemetry.instrumentation.httpx import HTTPXClientInstrumentor

# Create a single async client instance for use across
# the SDK.
# Each instance creates its own connection pool, so it's best
# to re-use them if possible.
# See: https://www.python-httpx.org/async/#opening-and-closing-clients
async_client = AsyncClient()



from e80_sdk.internal.environment import Environment, UserApiKey, JobToken
from e80_sdk.internal.httpx_async import async_client
from pydantic import BaseModel


class PlatformClient:
    _env: Environment

    def __init__(self, env: Environment):
        self._env = env

    async def create_sandbox(self) -> "CreateSandboxResponse":
        headers = {}
        if isinstance(self._env.identity, UserApiKey):
            headers["authorization"] = f"Bearer {self._env.identity.api_key}"
        elif isinstance(self._env.identity, JobToken):
            headers["authorization"] = f"Bearer {self._env.identity.job_token}"
            headers["x-8080-job-token"] = "1"

        resp = await async_client.post(
            f"{self._env.base_platform_url}/api/sandbox/{self._env.organization_slug}/{self._env.project_slug}/deploy",
            headers=headers,
        )
        resp.raise_for_status()
        return CreateSandboxResponse.model_validate(resp.json())


class CreateSandboxResponse(BaseModel):
    address: str
    auth_token: str



import os
import openai

from typing import AsyncIterator, overload, Optional
from contextlib import asynccontextmanager
from e80_sdk.internal.platform import PlatformClient
from e80_sdk.errors import Eighty80FatalError
from e80_sdk.secrets import Secrets
from e80_sdk.sandbox import SandboxClient
from e80_sdk.internal.environment import Environment, UserApiKey, JobToken
from openai.resources.responses.responses import Responses, AsyncResponses
from openai.resources.chat.chat import Chat, AsyncChat
from openai.resources.models import Models, AsyncModels
from opentelemetry.instrumentation.httpx import HTTPXClientInstrumentor

# Instrument all HTTPX clients with OpenTelemetry
# Since we do not manually pass in clients into OpenAI SDK,
# we simply do this to get the calls to OpenAI instrumented.
HTTPXClientInstrumentor().instrument()


class Eighty80:
    """
    The main entrypoint into the 8080 SDK.

    Call its methods to access different 8080 services.
    """

    _cached: tuple[Secrets, Environment] | None = None
    _environment: Environment
    _secrets: Secrets

    def __init__(
        self, *, env: Optional[Environment] = None, secrets: Optional[Secrets] = None
    ):
        if env is not None and secrets is not None:
            self._environment = env
            self._secrets = secrets
        # Use a cached version of the environment and secrets if called with
        # no arguments.
        elif Eighty80._cached:
            self._secrets, self._environment = Eighty80._cached
        else:
            load = self._load_environment()
            Eighty80._cached = load
            self._secrets, self._environment = load

    def secrets(self) -> Secrets:
        return self._secrets

    @overload
    def completion_sdk(self) -> "_Eighty80OpenAISDK": ...

    @overload
    def completion_sdk(self, secret_name: str) -> openai.OpenAI: ...

    def completion_sdk(self, secret_name: str | None = None):
        """
        Obtain an OpenAI SDK-compatible object that can be a drop-in
        replacement for the OpenAI SDK client.

        ```python
        from openai import OpenAI
        from e80_sdk import Eighty80

        old_client = OpenAI(...)
        drop_me_in = Eighty80.completion_sdk(...)
        ```

        To use 8080 models, simply call this method without any arguments.
        Note that not all properties are implemented for 8080 models.

        ```python
        from e80_sdk import Eighty80

        Eighty80.completion_sdk().responses.create(...)
        Eighty80.completion_sdk().chat.completions.create(...)
        ```

        You can set an OpenAI SDK secret in the 8080 platform, and obtain
        an OpenAI SDK-compatible object. An SDK object created this way has
        no restrictions on what properties are available.

        ```python
        from e80_sdk import Eighty80

        # Assumes 'foo' is an OpenAI SDK secret set in the 8080 platform
        Eighty80.completion_sdk('foo').chat.completions.create(...)
        ```
        """
        if secret_name is None:
            return _Eighty80OpenAISDK(self._environment)

        secret = self.secrets().get_openai_secret(secret_name)
        return openai.OpenAI(base_url=secret.url, api_key=secret.api_key)

    @overload
    def async_completion_sdk(self) -> "_Eighty80OpenAIAsyncSDK": ...

    @overload
    def async_completion_sdk(self, secret_name: str) -> openai.AsyncOpenAI: ...

    def async_completion_sdk(self, secret_name: str | None = None):
        """
        Obtain an async OpenAI SDK-compatible object that can be a drop-in
        replacement for the async OpenAI SDK client.

        ```python
        from openai import AsyncOpenAI
        from e80_sdk import Eighty80

        old_client = AsyncOpenAI(...)
        drop_me_in = Eighty80.async_completion_sdk(...)
        ```

        To use 8080 models, simply call this method without any arguments.
        Note that not all properties are implemented for 8080 models.

        ```python
        from e80_sdk import Eighty80

        await Eighty80.async_completion_sdk().responses.create(...)
        await Eighty80.async_completion_sdk().chat.completions.create(...)
        ```

        You can set an OpenAI SDK secret in the 8080 platform, and obtain
        an async OpenAI SDK-compatible object. An SDK object created this way has
        no restrictions on what properties are available.

        ```python
        from e80_sdk import Eighty80

        # Assumes 'foo' is an OpenAI SDK secret set in the 8080 platform
        await Eighty80.async_completion_sdk('foo').chat.completions.create(...)
        ```
        """

        if secret_name is None:
            return _Eighty80OpenAIAsyncSDK(self._environment)

        secret = self.secrets().get_openai_secret(secret_name)
        return openai.AsyncOpenAI(base_url=secret.url, api_key=secret.api_key)

    @asynccontextmanager
    async def sandbox(self) -> AsyncIterator[SandboxClient]:
        """
        Create a new sandbox where you can execute arbitrary Python
        and Javascript code.

        Use this as an async context manager so the sandbox will get cleaned up
        automatically.

        ```
        from e80_sdk import Eighty80

        async with Eighty80.sandbox() as sandbox_client:
            sandbox_client.run_python("print('hello world')")
            sandbox_client.install_python_dependencies(["requests"])
        ```
        """
        client = PlatformClient(self._environment)

        sb_resp = await client.create_sandbox()
        sb_client = SandboxClient(sb_resp.auth_token, sb_resp.address)
        sb_client.wait_until_ready()

        try:
            yield sb_client
        finally:
            sb_client.destroy()

    def _load_environment(self) -> tuple[Secrets, Environment]:
        """
        Loads the 8080 environment from environment variables and Nomad files.

        Basically, this is used for loading jobs running in prod.

        Note, the environment can be populated manually via the constructor.
        More options are available when creating the environment manually.
        """

        # Required env variables - MUST be populated no matter what.
        project_slug = os.environ["8080_PROJECT_SLUG"]
        org_slug = os.environ["8080_ORGANIZATION_SLUG"]

        # Optional env variables
        api_url = os.environ.get("8080_API_URL", "https://api.8080.io")
        platform_url = os.environ.get("8080_PLATFORM_URL", "https://app.8080.io")
        use_secrets_file = os.environ.get("8080_SECRETS_FILE", None)
        secrets_json = os.environ.get("8080_SECRETS_JSON", None)

        # One of these MUST be set.
        identity_token = os.environ.get("8080_IDENTITY_TOKEN", None)
        user_api_key = os.environ.get("8080_API_KEY", None)

        identity: JobToken | UserApiKey | None = None
        if identity_token:
            identity = JobToken(job_token=identity_token)
        elif user_api_key:
            identity = UserApiKey(api_key=user_api_key)
        if identity is None:
            raise Eighty80FatalError(
                "Identity was None. Did you populate 8080_IDENTITY_TOKEN or 8080_API_KEY?"
            )

        return Secrets(
            secrets_file="/local/secrets.json" if use_secrets_file else None,
            secrets_json=secrets_json,
        ), Environment(
            organization_slug=org_slug,
            project_slug=project_slug,
            identity=identity,
            base_platform_url=platform_url,
            base_api_url=api_url,
        )


class _Eighty80OpenAISDK:
    _cached_client: openai.OpenAI

    def __init__(self, env: Environment):
        client = None

        if isinstance(env.identity, JobToken):
            client = openai.OpenAI(
                api_key=env.identity.job_token,
                base_url=f"{env.base_api_url}/v1",
                default_headers={"x-8080-job-token": "1"},
            )
        elif isinstance(env.identity, UserApiKey):
            client = openai.OpenAI(
                api_key=env.identity.api_key,
                base_url=f"{env.base_api_url}/v1",
                default_headers={"x-8080-project-slug": env.project_slug},
            )
        if client is None:
            raise Exception("Somehow got a null client")
        self._cached_client = client

    @property
    def chat(self) -> Chat:
        return self._cached_client.chat

    @property
    def models(self) -> Models:
        return self._cached_client.models

    @property
    def responses(self) -> Responses:
        return self._cached_client.responses


class _Eighty80OpenAIAsyncSDK:
    _cached_client: openai.AsyncOpenAI

    def __init__(self, env: Environment):
        client = None
        if isinstance(env.identity, JobToken):
            client = openai.AsyncOpenAI(
                api_key=env.identity.job_token,
                base_url=f"{env.base_api_url}/v1",
                default_headers={"x-8080-job-token": "1"},
            )
        elif isinstance(env.identity, UserApiKey):
            client = openai.AsyncOpenAI(
                api_key=env.identity.api_key,
                base_url=f"{env.base_api_url}/v1",
                default_headers={"x-8080-project-slug": env.project_slug},
            )
        if client is None:
            raise Exception("Somehow got a null client")
        self._cached_client = client

    @property
    def chat(self) -> AsyncChat:
        return self._cached_client.chat

    @property
    def models(self) -> AsyncModels:
        return self._cached_client.models

    @property
    def responses(self) -> AsyncResponses:
        return self._cached_client.responses



from dataclasses import dataclass


@dataclass
class UserApiKey:
    api_key: str


@dataclass
class JobToken:
    job_token: str


@dataclass
class Environment:
    organization_slug: str
    project_slug: str
    identity: UserApiKey | JobToken
    base_platform_url: str
    base_api_url: str



import httpx
import time
from pydantic import BaseModel

from dataclasses import dataclass


class SandboxClient:
    _auth_token: str
    _base_url: str

    def __init__(self, auth_token: str, base_url: str):
        self._auth_token = auth_token
        self._base_url = base_url
        self.session = httpx.Client(
            headers={
                "Authorization": f"Bearer {self._auth_token}",
                "Content-Type": "application/json",
            }
        )

    def wait_until_ready(self):
        for _ in range(10):
            try:
                resp = self.session.get(f"{self._base_url}/.8080/health")
                resp.raise_for_status()
                return
            except Exception:
                pass
            time.sleep(1)

    def run_python(self, code: str) -> "RunCodeResult":
        return self._run_command("python", code)

    def run_javascript(self, code: str) -> "RunCodeResult":
        return self._run_command("js", code)

    def install_javascript_dependency(self, deps: list[str]):
        return self._install_dependencies("js", deps)

    def install_python_dependencies(self, deps: list[str]):
        return self._install_dependencies("python", deps)

    def _run_command(self, lang: str, code: str) -> "RunCodeResult":
        resp = self.session.post(f"{self._base_url}/{lang}/run", json={"code": code})
        resp.raise_for_status()
        run_code_resp = _RunCodeResponse.model_validate(resp.json())
        return RunCodeResult(
            is_successful=run_code_resp.exit_code == 0,
            result=run_code_resp.message,
            error=run_code_resp.error,
        )

    def _install_dependencies(self, lang: str, deps: list[str]):
        resp = self.session.post(
            f"{self._base_url}/{lang}/deps", json={"dependencies": deps}
        )
        resp.raise_for_status()

    def destroy(self):
        self.session.delete(f"{self._base_url}/me")


@dataclass
class RunCodeResult:
    is_successful: bool
    result: str
    error: str


class _RunCodeResponse(BaseModel):
    exit_code: int
    message: str
    error: str



from e80_sdk import Eighty80


def generate(
    prompt: str,
    model: str = "8080/8080",
    stream: bool = False,
):
    client = Eighty80().completion_sdk()
    messages = [{"role": "user", "content": prompt}]

    response = client.chat.completions(messages, model=model, stream=stream)

    content = response["choices"][0]["message"]["content"]
    return content



# adding common functions to base path so that:
# `from e80_sdk import eighty80_app, generate`
# imports above will work, without needing to import common functions from the submodules
# making for example SDK code easier to write and read
from .endpoint import eighty80_app
from .eighty80 import Eighty80
from .generate import generate



import logging
from . import fastapi
from e80_sdk.internal.httpx_async import async_client
from contextlib import asynccontextmanager
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

logger = logging.getLogger(__name__)

reload_jwt_task = None

_EIGHTY80_HEALTH_CHECK_PATH = "/.8080/health"


def eighty80_app() -> fastapi.FastAPI:
    global reload_jwt_task  # TODO: Remove this global.

    eighty80_app = fastapi.FastAPI()
    FastAPIInstrumentor.instrument_app(
        eighty80_app,
        excluded_urls=_EIGHTY80_HEALTH_CHECK_PATH,
    )

    eighty80_app.get(_EIGHTY80_HEALTH_CHECK_PATH)(_eighty80_health_check)

    return eighty80_app


async def _eighty80_health_check():
    return {"status": "ok"}


@asynccontextmanager
async def lifespan(app: fastapi.FastAPI):
    yield
    await async_client.aclose()
